import time
import warnings
from abc import ABC, abstractmethod
import torch
import numpy as np
import SimpleITK as sitk
from ..utils import is_image_dict, nib_to_sitk, sitk_to_nib


class Transform(ABC):
    def __init__(self, verbose=False):
        self.verbose = verbose

    def __call__(self, sample):
        if self.verbose:
            start = time.time()
        self.parse_sample(sample)
        sample = self.apply_transform(sample)
        if self.verbose:
            duration = time.time() - start
            print(f'{self.__class__.__name__}: {duration:.3f} seconds')
        return sample

    @abstractmethod
    def apply_transform(self, sample):
        pass

    @staticmethod
    def parse_sample(sample):
        images_found = False
        type_in_dict = False
        for image_dict in sample.values():
            if not is_image_dict(image_dict):
                continue
            images_found = True
            if 'type' in image_dict:
                type_in_dict = True
            if images_found and type_in_dict:
                break
        if not images_found:
            warnings.warn(
                'No image dicts found in sample.'
                f' Sample keys: {sample.keys()}'
            )

    @staticmethod
    def nib_to_sitk(data, affine):
        return nib_to_sitk(data, affine)

    @staticmethod
    def sitk_to_nib(image):
        return sitk_to_nib(image)



import torch
import numpy as np
import nibabel as nib
from ....utils import is_image_dict
from ....torchio import DATA, AFFINE
from ... import Transform


class ToCanonical(Transform):
    def __init__(
            self,
            verbose=False,
            ):
        """Reorder the data to be closest to canonical (RAS+) orientation"""
        super().__init__(verbose=verbose)

    def apply_transform(self, sample):
        for image_dict in sample.values():
            if not is_image_dict(image_dict):
                continue
            affine = image_dict[AFFINE]
            if nib.aff2axcodes(affine) == tuple('RAS'):
                continue
            array = image_dict[DATA][0].numpy()
            nii = nib.Nifti1Image(array, affine)
            reoriented = nib.as_closest_canonical(nii)
            array = reoriented.get_fdata(dtype=np.float32)
            # https://github.com/facebookresearch/InferSent/issues/99#issuecomment-446175325
            array = array.copy()[np.newaxis, ...]
            image_dict[DATA] = torch.from_numpy(array)
            image_dict[AFFINE] = reoriented.affine
        return sample



"""Main module."""

__all__ = [
    'INTENSITY',
    'LABEL',
    'SAMPLING_MAP',
    'DATA',
    'AFFINE',
    'IMAGE',
    'LOCATION',
]

INTENSITY = 'intensity'
LABEL = 'label'
SAMPLING_MAP = 'sampling_map'

DATA = 'data'
AFFINE = 'affine'

# For aggregator
IMAGE = 'image'
LOCATION = 'location'



import numbers
from abc import abstractmethod
import torch
from .. import Transform, Interpolation


class RandomTransform(Transform):
    def __init__(self, seed=None, verbose=False):
        super().__init__(verbose=verbose)
        self.seed = seed

    def __call__(self, sample):
        self.check_seed()
        return super().__call__(sample)

    @staticmethod
    @abstractmethod
    def get_params(*args, **kwargs):
        pass

    @staticmethod
    def parse_range(nums_range, name):
        """Adapted from torchvision.RandomRotation"""
        if isinstance(nums_range, numbers.Number):
            if nums_range < 0:
                raise ValueError(
                    f'If {name} is a single number,'
                    f' it must be positive, not {nums_range}')
            return (-nums_range, nums_range)
        else:
            if len(nums_range) != 2:
                raise ValueError(
                    f'If {name} is a sequence,'
                    f' it must be of len 2, not {nums_range}')
            min_degree, max_degree = nums_range
            if min_degree > max_degree:
                raise ValueError(
                    f'If {name} is a sequence, the second value must be'
                    f' equal or greater than the first, not {nums_range}')
            return nums_range

    def parse_degrees(self, degrees):
        return self.parse_range(degrees, 'degrees')

    def parse_translation(self, translation):
        return self.parse_range(translation, 'translation')

    @staticmethod
    def parse_probability(p, name):
        if not (isinstance(p, numbers.Number) and 0 <= p <= 1):
            raise ValueError(f'{name} must be a number in [0, 1]')
        return p

    @staticmethod
    def parse_interpolation(interpolation):
        if not isinstance(interpolation, Interpolation):
            message = (
                'image_interpolation must be'
                ' a member of torchio.Interpolation'
            )
            raise TypeError(message)
        return interpolation

    def check_seed(self):
        if self.seed is not None:
            if self.verbose:
                print('Setting torch seed to', self.seed)
            torch.manual_seed(self.seed)



import torch
import numpy as np
import nibabel as nib
from nibabel.processing import resample_to_output
from ....utils import is_image_dict
from ....torchio import LABEL, DATA, AFFINE
from ... import Interpolation
from ... import Transform


class Resample(Transform):
    def __init__(
            self,
            voxel_sizes,
            antialiasing=True,
            image_interpolation=Interpolation.LINEAR,
            verbose=False,
            ):
        super().__init__(verbose=verbose)
        self.voxel_sizes = voxel_sizes
        self.antialiasing = antialiasing
        self.image_interpolation = image_interpolation

    def apply_transform(self, sample):
        for image_dict in sample.values():
            if not is_image_dict(image_dict):
                continue
            if image_dict['type'] == LABEL:
                interpolation = Interpolation.NEAREST
            else:
                interpolation = self.image_interpolation
            image_dict[DATA], image_dict[AFFINE] = self.apply_resample(
                image_dict[DATA],
                image_dict[AFFINE],
                self.voxel_sizes,
                interpolation,
            )
        return sample

    def apply_resample(
            self,
            tensor,
            affine,
            voxel_sizes,
            interpolation: Interpolation,
            ):
        if interpolation == Interpolation.NEAREST:
            order = 0
        elif interpolation == Interpolation.LINEAR:
            order = 1
        elif interpolation == Interpolation.BSPLINE:
            order = 3
        else:
            message = f'Interpolation not implemented yet: {interpolation}'
            raise NotImplementedError(message)
        array = tensor.numpy()[0]
        nii = resample_to_output(
            nib.Nifti1Image(array, affine),
            voxel_sizes=voxel_sizes,
            order=order,
        )
        tensor = torch.from_numpy(nii.get_fdata(dtype=np.float32))
        tensor = tensor.unsqueeze(dim=0)
        return tensor, nii.affine



import random
import warnings
from itertools import islice
from tqdm import trange
from torch.utils.data import Dataset, DataLoader


class Queue(Dataset):
    def __init__(
            self,
            subjects_dataset,
            max_length,
            samples_per_volume,
            patch_size,
            sampler_class,
            num_workers=0,
            shuffle_subjects=True,
            shuffle_patches=True,
            verbose=False,
            ):
        self.subjects_dataset = subjects_dataset
        self.max_length = max_length
        self.shuffle_subjects = shuffle_subjects
        self.shuffle_patches = shuffle_patches
        self.samples_per_volume = samples_per_volume
        self.sampler_class = sampler_class
        self.patch_size = patch_size
        self.num_workers = num_workers
        self.verbose = verbose
        self.subjects_iterable = self.get_subjects_iterable()
        self.patches_list = []
        self.num_sampled_patches = 0

    def __len__(self):
        return self.iterations_per_epoch

    def __getitem__(self, _):
        """There are probably more elegant ways of doing this"""
        if not self.patches_list:
            self.print('Patches list is empty.')
            self.fill()
        sample_patch = self.patches_list.pop()
        self.num_sampled_patches += 1
        return sample_patch

    def __repr__(self):
        attributes = [
            f'max_length={self.max_length}',
            f'num_subjects={self.num_subjects}',
            f'num_patches={self.num_patches}',
            f'samples_per_volume={self.samples_per_volume}',
            f'num_sampled_patches={self.num_sampled_patches}',
            f'iterations_per_epoch={self.iterations_per_epoch}',
        ]
        attributes_string = ', '.join(attributes)
        return f'Queue({attributes_string})'

    def print(self, *args):
        if self.verbose:
            print(*args)

    @property
    def num_subjects(self):
        return len(self.subjects_dataset)

    @property
    def num_patches(self):
        return len(self.patches_list)

    @property
    def iterations_per_epoch(self):
        return self.num_subjects * self.samples_per_volume

    def fill(self):
        assert self.sampler_class is not None
        assert self.patch_size is not None
        if self.max_length % self.samples_per_volume != 0:
            message = (
                f'Samples per volume ({self.samples_per_volume})'
                f' not divisible by max length ({self.max_length})'
            )
            warnings.warn(message)

        # If there are e.g. 4 subjects and 1 sample per volume and max_length
        # is 6, we just need to load 4 subjects, not 6
        max_num_subjects_for_queue = self.max_length // self.samples_per_volume
        num_subjects_for_queue = min(
            self.num_subjects, max_num_subjects_for_queue)

        self.print(f'Filling queue from {num_subjects_for_queue} subjects...')
        if self.verbose:
            iterable = trange(num_subjects_for_queue, leave=False)
        else:
            iterable = range(num_subjects_for_queue)
        for _ in iterable:
            subject_sample = self.get_next_subject_sample()
            sampler = self.sampler_class(subject_sample, self.patch_size)
            samples = [s for s in islice(sampler, self.samples_per_volume)]
            assert isinstance(samples, list)
            self.patches_list.extend(samples)
        if self.shuffle_patches:
            random.shuffle(self.patches_list)

    def get_next_subject_sample(self):
        """A StopIteration exception is expected when the queue is empty"""
        try:
            subject_sample = next(self.subjects_iterable)
        except StopIteration as exception:
            self.print('Queue is empty:', exception)
            self.subjects_iterable = self.get_subjects_iterable()
            subject_sample = next(self.subjects_iterable)
        return subject_sample

    def get_subjects_iterable(self):
        """
        I need a DataLoader to handle parallelism
        But this loader is always expected to yield single subject samples
        """
        self.print(
            '\nCreating subjects loader with', self.num_workers, 'workers')
        subjects_loader = DataLoader(
            self.subjects_dataset,
            num_workers=self.num_workers,
            collate_fn=lambda x: x[0],
            shuffle=self.shuffle_subjects,
        )
        return iter(subjects_loader)



import numpy as np
from torchio.utils import to_tuple


class GridAggregator:
    """
    Adapted from NiftyNet.
    See https://niftynet.readthedocs.io/en/dev/window_sizes.html
    """
    def __init__(self, data, patch_overlap):
        self.output_array = np.full(
            data.shape,
            fill_value=0,
            dtype=np.uint16,
        )
        self.patch_overlap = to_tuple(patch_overlap)

    @staticmethod
    def crop_batch(windows, location, border):
        location = location.astype(np.int)
        batch_shape = windows.shape
        spatial_shape = batch_shape[2:]  # ignore batch and channels dim
        num_dimensions = 3
        for idx in range(num_dimensions):
            location[:, idx] = location[:, idx] + border[idx]
            location[:, idx + 3] = location[:, idx + 3] - border[idx]
        cropped_shape = np.max(location[:, 3:6] - location[:, 0:3], axis=0)
        diff = spatial_shape - cropped_shape
        left = np.floor(diff / 2).astype(np.int)
        i_ini, j_ini, k_ini = left
        i_fin, j_fin, k_fin = left + cropped_shape
        batch = windows[
            :,  # batch dimension
            :,  # channels dimension
            i_ini:i_fin,
            j_ini:j_fin,
            k_ini:k_fin,
        ]
        return batch, location

    def add_batch(self, windows, locations):
        windows = windows.cpu()
        location_init = np.copy(locations)
        init_ones = np.ones_like(windows)
        windows, _ = self.crop_batch(
            windows, location_init,
            self.patch_overlap,
        )
        location_init = np.copy(locations)
        _, locations = self.crop_batch(
            init_ones,
            location_init,
            self.patch_overlap,
        )
        for window, location in zip(windows, locations):
            window = window[0]
            i_ini, j_ini, k_ini, i_fin, j_fin, k_fin = location
            self.output_array[i_ini:i_fin, j_ini:j_fin, k_ini:k_fin] = window



from .sampler import ImageSampler


class LabelSampler(ImageSampler):
    """
    This iterable dataset yields patches that contain at least one voxel
    without background. See main() for an example.

    For now, this implementation is not efficient because it uses brute force
    to look for foreground voxels.
    """
    # pylint: disable=abstract-method
    def extract_patch_generator(self, sample, patch_size):
        while True:
            yield self.extract_patch(sample, patch_size)

    def extract_patch(self, sample, patch_size):
        has_label = False
        while not has_label:
            index_ini, index_fin = self.get_random_indices(
                sample, patch_size)
            patch_label = self.crop(sample['label'], index_ini, index_fin)
            has_label = patch_label.sum() > 0
        cropped_sample = self.copy_and_crop(
            sample,
            index_ini,
            index_fin,
        )
        return cropped_sample



import torch
from ....utils import is_image_dict
from ....torchio import DATA, INTENSITY
from ... import Transform


class NormalizationTransform(Transform):
    def __init__(self, masking_method=None, verbose=False):
        """
        masking_method is used to choose the values used for normalization.
        It can be:
         - A string: the mask will be retrieved from the sample
         - A function: the mask will be computed using the function
         - None: all values are used
        """
        super().__init__(verbose=verbose)
        self.mask_name = None
        if masking_method is None:
            self.masking_method = self.ones
        elif callable(masking_method):
            self.masking_method = masking_method
        elif isinstance(masking_method, str):
            self.mask_name = masking_method

    def get_mask(self, sample, data):
        if self.mask_name is None:
            return self.masking_method(data)
        else:
            return sample[self.mask_name][DATA].bool()

    def apply_transform(self, sample):
        for image_name, image_dict in sample.items():
            if not is_image_dict(image_dict):
                continue
            if not image_dict['type'] == INTENSITY:
                continue
            mask = self.get_mask(sample, image_dict[DATA])
            self.apply_normalization(sample, image_name, mask)
        return sample

    def apply_normalization(self, sample, image_name, mask):
        """There must be a nicer way of doing this"""
        raise NotImplementedError

    @staticmethod
    def ones(data):
        return torch.ones_like(data, dtype=torch.bool)

    @staticmethod
    def mean(data):
        mask = data > data.mean()
        return mask



import torch
import numpy as np
import SimpleITK as sitk
from ....utils import is_image_dict, check_consistent_shape
from ....torchio import LABEL, DATA, AFFINE
from .. import Interpolation
from .. import RandomTransform


class RandomAffine(RandomTransform):
    def __init__(
            self,
            scales=(0.9, 1.1),
            degrees=10,
            isotropic=False,
            image_interpolation=Interpolation.LINEAR,
            seed=None,
            verbose=False,
            ):
        super().__init__(seed=seed, verbose=verbose)
        self.scales = scales
        self.degrees = self.parse_degrees(degrees)
        self.isotropic = isotropic
        self.interpolation = self.parse_interpolation(image_interpolation)

    def apply_transform(self, sample):
        check_consistent_shape(sample)
        scaling_params, rotation_params = self.get_params(
            self.scales, self.degrees, self.isotropic)
        sample['random_scaling'] = scaling_params
        sample['random_rotation'] = rotation_params
        for image_dict in sample.values():
            if not is_image_dict(image_dict):
                continue
            if image_dict['type'] == LABEL:
                interpolation = Interpolation.NEAREST
            else:
                interpolation = self.interpolation
            image_dict[DATA] = self.apply_affine_transform(
                image_dict[DATA],
                image_dict[AFFINE],
                scaling_params,
                rotation_params,
                interpolation,
            )
        return sample

    @staticmethod
    def get_params(scales, degrees, isotropic):
        scaling_params = torch.FloatTensor(3).uniform_(*scales)
        if isotropic:
            scaling_params.fill_(scaling_params[0])
        rotation_params = torch.FloatTensor(3).uniform_(*degrees)
        return scaling_params.tolist(), rotation_params.tolist()

    @staticmethod
    def get_scaling_transform(scaling_params):
        """
        scaling_params are inverted so that they are more intuitive
        For example, 1.5 means the objects look 1.5 times larger
        """
        transform = sitk.ScaleTransform(3)
        scaling_params = 1 / np.array(scaling_params)
        transform.SetScale(scaling_params)
        return transform

    @staticmethod
    def get_rotation_transform(degrees):
        transform = sitk.Euler3DTransform()
        radians = np.radians(degrees)
        transform.SetRotation(*radians)
        return transform

    def apply_affine_transform(
            self,
            tensor,
            affine,
            scaling_params,
            rotation_params,
            interpolation: Interpolation,
            ):
        assert tensor.ndim == 4
        assert len(tensor) == 1

        image = self.nib_to_sitk(tensor[0], affine)
        floating = reference = image

        scaling_transform = self.get_scaling_transform(scaling_params)
        rotation_transform = self.get_rotation_transform(rotation_params)
        transform = sitk.Transform(3, sitk.sitkComposite)
        transform.AddTransform(scaling_transform)
        transform.AddTransform(rotation_transform)

        resampler = sitk.ResampleImageFilter()
        resampler.SetInterpolator(interpolation.value)
        resampler.SetReferenceImage(reference)
        resampler.SetDefaultPixelValue(tensor.min().item())
        resampler.SetOutputPixelType(sitk.sitkFloat32)
        resampler.SetTransform(transform)
        resampled = resampler.Execute(floating)

        np_array = sitk.GetArrayFromImage(resampled)
        np_array = np_array.transpose()  # ITK to NumPy
        tensor[0] = torch.from_numpy(np_array)
        return tensor



from ....torchio import DATA
from .normalization_transform import NormalizationTransform


class ZNormalization(NormalizationTransform):
    """
    Subtract mean and divide by standard deviation
    """
    def __init__(self, masking_method=None, verbose=False):
        super().__init__(masking_method=masking_method, verbose=verbose)

    def apply_normalization(self, sample, image_name, mask):
        image_dict = sample[image_name]
        image_dict[DATA] = self.znorm(
            image_dict[DATA],
            mask,
        )

    @staticmethod
    def znorm(data, mask):
        values = data[mask]
        mean, std = values.mean(), values.std()
        data = data - mean
        data = data / std
        return data



import warnings
import torch
import numpy as np
from ....torchio import DATA
from .normalization_transform import NormalizationTransform


class Rescale(NormalizationTransform):
    def __init__(
            self,
            out_min_max,
            percentiles=(0, 100),
            masking_method=None,
            verbose=False,
            ):
        super().__init__(masking_method=masking_method, verbose=verbose)
        self.out_min, self.out_max = out_min_max
        self.percentiles = percentiles

    def apply_normalization(self, sample, image_name, mask):
        image_dict = sample[image_name]
        image_dict[DATA] = self.rescale(image_dict[DATA], mask, image_name)

    def rescale(self, data, mask, image_name):
        array = data.numpy()
        mask = mask.numpy()
        values = array[mask]
        cutoff = np.percentile(values, self.percentiles)
        np.clip(array, *cutoff, out=array)
        array -= array.min()  # [0, max]
        array_max = array.max()
        if array_max == 0:
            message = (
                f'Rescaling image "{image_name}" not possible'
                ' due to division by zero'
            )
            warnings.warn(message)
            return data
        array /= array.max()  # [0, 1]
        out_range = self.out_max - self.out_min
        array *= out_range  # [0, out_range]
        array += self.out_min  # [out_min, out_max]
        return torch.from_numpy(array)



import SimpleITK as sitk
from .bounds_transform import BoundsTransform


class Crop(BoundsTransform):
    def __init__(
            self,
            cropping,
            verbose=False,
            ):
        """
        cropping should be an integer or a tuple of 3 or 6 integers
        """
        super().__init__(cropping, verbose=verbose)

    @property
    def bounds_function(self):
        return sitk.Crop



"""Top-level package for torchio."""

__author__ = """Fernando Perez-Garcia"""
__email__ = 'fernando.perezgarcia.17@ucl.ac.uk'
__version__ = '0.12.8'

from . import utils
from .torchio import *
from .transforms import *
from .data import io, sampler, inference, ImagesDataset, Image, Queue, Subject



import enum
import SimpleITK as sitk


@enum.unique
class Interpolation(enum.IntEnum):
    NEAREST = sitk.sitkNearestNeighbor
    LINEAR = sitk.sitkLinear
    BSPLINE = sitk.sitkBSpline
    GAUSSIAN = sitk.sitkGaussian
    LABEL_GAUSSIAN = sitk.sitkLabelGaussian
    HAMMING = sitk.sitkHammingWindowedSinc
    COSINE = sitk.sitkCosineWindowedSinc
    WELCH = sitk.sitkWelchWindowedSinc
    LANCZOS = sitk.sitkLanczosWindowedSinc
    BLACKMAN = sitk.sitkBlackmanWindowedSinc



import SimpleITK as sitk
from .bounds_transform import BoundsTransform


class Pad(BoundsTransform):
    PADDING_FUNCTIONS = {
        'constant': sitk.ConstantPad,
        'reflect': sitk.MirrorPad,
        'replicate': sitk.ZeroFluxNeumannPad,
        'circular': sitk.WrapPad,
        }

    def __init__(
            self,
            padding,
            padding_mode='constant',
            fill=None,
            verbose=False,
            ):
        """
        padding_mode can be 'constant', 'reflect', 'replicate' or 'circular'.
        See https://pytorch.org/docs/stable/nn.functional.html#pad for more
        information about this transform.
        """
        super().__init__(padding, verbose=verbose)
        self.padding_mode = padding_mode
        self.fill = fill

    @property
    def bounds_function(self):
        try:
            return self.PADDING_FUNCTIONS[self.padding_mode]
        except KeyError:
            message = (
                f'padding_mode "{self.padding_mode}" not valid.'
                f' Valid options are {list(self.PADDING_FUNCTIONS.keys())}'
            )
            raise ValueError(message)
